from dataclasses import dataclass
from ipaddress import IPv4Address, AddressValueError
from typing import Optional

import pytest

from serpyco_rs import Serializer, SchemaValidationError, ErrorItem
from serpyco_rs._custom_types import CustomType


class IPv4AddressType(CustomType[IPv4Address, str]):
    def serialize(self, value: IPv4Address) -> str:
        return str(value)

    def deserialize(self, value: str) -> IPv4Address:
        return IPv4Address(value)

    def get_json_schema(self):
        return {
            'type': 'string',
            'format': 'ipv4',
        }


def custom_type_resolver(t: type) -> Optional[CustomType]:
    if t is IPv4Address:
        return IPv4AddressType()
    return None


def test_custom_type():
    @dataclass
    class Data:
        ip: IPv4Address

    data = Data(ip=IPv4Address('1.1.1.1'))

    serializer = Serializer(Data, custom_type_resolver=custom_type_resolver)

    assert serializer.dump(data) == {'ip': '1.1.1.1'}
    assert serializer.load({'ip': '1.1.1.1'}) == data
    assert serializer.get_json_schema() == {
        '$ref': '#/components/schemas/tests.test_custom_types.test_custom_type.<locals>.Data[no_format,keep_nones]',
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        'components': {
            'schemas': {
                'tests.test_custom_types.test_custom_type.<locals>.Data[no_format,keep_nones]': {
                    'properties': {
                        'ip': {
                            'format': 'ipv4',
                            'type': 'string',
                        }
                    },
                    'required': ['ip'],
                    'type': 'object',
                }
            }
        },
    }


def test_custom_type__validation_error():
    @dataclass
    class Data:
        ip: IPv4Address

    serializer = Serializer(Data, custom_type_resolver=custom_type_resolver)
    with pytest.raises(SchemaValidationError) as exc_info:
        serializer.load({'ip': 'invalid'})

    assert exc_info.value.errors == [
        ErrorItem(message="AddressValueError: Expected 4 octets in 'invalid'", instance_path='ip')
    ]
    assert isinstance(exc_info.value.__cause__, AddressValueError)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from ._consts import DEFAULT_REF_PREFIX


@dataclass
class Config:
    ref_prefix: str = DEFAULT_REF_PREFIX


@dataclass
class Schema:
    config: Config

    type: str | None = None
    title: str | None = None
    description: str | None = None
    default: Any | None = None
    enum: list[Any] | None = None

    allOf: list[Schema] | None = None
    anyOf: list[Schema] | None = None
    oneOf: list[Schema] | None = None
    additionalArgs: dict[str, Any] | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = {
            'type': self.type,
            'title': self.title,
            'description': self.description,
            'default': self.default,
            'enum': self.enum,
            'allOf': [item.dump(definitions) for item in self.allOf] if self.allOf else None,
            'anyOf': [item.dump(definitions) for item in self.anyOf] if self.anyOf else None,
            'oneOf': [item.dump(definitions) for item in self.oneOf] if self.oneOf else None,
            **(self.additionalArgs or {}),
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class Boolean(Schema):
    type: str = 'boolean'  # pyright: ignore[reportIncompatibleVariableOverride]


@dataclass
class Null(Schema):
    type: str = 'null'  # pyright: ignore[reportIncompatibleVariableOverride]


@dataclass
class StringType(Schema):
    type: str = 'string'  # pyright: ignore[reportIncompatibleVariableOverride]
    minLength: int | None = None
    maxLength: int | None = None
    format: str | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            'minLength': self.minLength,
            'maxLength': self.maxLength,
            'format': self.format,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class NumberType(Schema):
    type: str = 'number'  # pyright: ignore[reportIncompatibleVariableOverride]
    minimum: float | None = None
    maximum: float | None = None
    format: str | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            'minimum': self.minimum,
            'maximum': self.maximum,
            'format': self.format,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class IntegerType(NumberType):
    type: str = 'integer'


@dataclass
class ObjectType(Schema):
    name: str | None = None
    type: str = 'object'  # pyright: ignore[reportIncompatibleVariableOverride]
    properties: dict[str, Schema] | None = None
    additionalProperties: bool | Schema | None = None
    required: list[str] | None = None

    @property
    def ref(self) -> str:
        return f'{self.config.ref_prefix}/{self.name}'

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            'properties': {k: v.dump(definitions) for k, v in self.properties.items()} if self.properties else None,
            'additionalProperties': (
                self.additionalProperties.dump(definitions)
                if isinstance(self.additionalProperties, Schema)
                else self.additionalProperties
            ),
            'required': self.required,
            **data,
        }
        data = {k: v for k, v in data.items() if v is not None}
        if not self.name:
            return data
        definitions[self.name] = data
        return {
            '$ref': self.ref,
        }


@dataclass
class ArrayType(Schema):
    type: str = 'array'  # pyright: ignore[reportIncompatibleVariableOverride]
    items: Schema | None = None
    prefixItems: list[Schema] | None = None
    minItems: int | None = None
    maxItems: int | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            'items': self.items.dump(definitions) if self.items else None,
            'prefixItems': [i.dump(definitions) for i in self.prefixItems] if self.prefixItems else None,
            'minItems': self.minItems,
            'maxItems': self.maxItems,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class RefType(Schema):
    name: str | None = None

    @property
    def ref(self) -> str:
        return f'{self.config.ref_prefix}/{self.name}'

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        return {
            '$ref': self.ref,
            **data,
        }


@dataclass
class DiscriminatedUnionType(Schema):
    discriminator: Discriminator | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        return {
            'discriminator': self.discriminator.dump() if self.discriminator else None,
            **data,
        }


@dataclass
class Discriminator:
    property_name: str
    mapping: dict[str, str]

    def dump(self) -> dict[str, Any]:
        return {
            'propertyName': self.property_name,
            'mapping': self.mapping,
        }



from typing import Any, Optional

from pydantic import BaseModel

from .base import make_test_object


class Nested(BaseModel):
    name: str


class Dataclass(BaseModel):
    name: str
    value: int
    f: float
    b: bool
    nest: list[Nested]
    many: list[int]
    option: Optional[str] = None


test_object = make_test_object(Dataclass, Nested)


def load(data: dict[str, Any]) -> Dataclass:
    return Dataclass(**data)


def dump(obj: Dataclass) -> dict[str, Any]:
    return obj.model_dump()



from typing import Generic, TypeVar

from ._serpyco_rs import (
    AnyType,
    ArrayType,
    BaseType,
    BooleanType,
    BytesType,
    CustomEncoder as _CustomEncoder,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue as _DefaultValue,
    DictionaryType,
    DiscriminatedUnionType,
    EntityField,
    EntityType,
    EnumType,
    ErrorItem,
    FloatType,
    IntegerType,
    LiteralType,
    OptionalType,
    RecursionHolder,
    SchemaValidationError,
    Serializer,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    UUIDType,
    ValidationError,
    CustomType,
)


_T = TypeVar('_T')
_I = TypeVar('_I')
_O = TypeVar('_O')


class CustomEncoder(_CustomEncoder, Generic[_I, _O]):
    """pyo3 doesn't support specifying concrete types for generic methods."""


class DefaultValue(_DefaultValue, Generic[_T]):
    """pyo3 doesn't support specifying concrete types for generic methods."""


NOT_SET = DefaultValue.none()



import re
from functools import lru_cache

from attributes_doc import get_attributes_doc as _get_attributes_doc


CAMELCASE_RE = re.compile(r'(?!^)_([a-zA-Z])')


def to_camelcase(s: str) -> str:
    return CAMELCASE_RE.sub(lambda m: m.group(1).upper(), s).rstrip('_')


get_attributes_doc = lru_cache(_get_attributes_doc)



import abc
from typing import Any, Generic, TypeVar


_I = TypeVar('_I')
_O = TypeVar('_O')


class CustomType(abc.ABC, Generic[_I, _O]):
    @abc.abstractmethod
    def serialize(self, value: _I) -> _O: ...

    @abc.abstractmethod
    def deserialize(self, value: _O) -> _I: ...

    @abc.abstractmethod
    def get_json_schema(self) -> dict[str, Any]: ...



import dataclasses
import sys
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, IntEnum
from typing import (
    Annotated,
    Any,
    ForwardRef,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
    get_origin,
    overload,
)
from uuid import UUID

from typing_extensions import NotRequired, Required, assert_never, get_args, is_typeddict

from ._custom_types import CustomType as CustomTypeMeta
from ._impl import (
    NOT_SET,
    AnyType,
    ArrayType,
    BaseType,
    BooleanType,
    BytesType,
    CustomEncoder,
    CustomType,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue,
    DictionaryType,
    DiscriminatedUnionType,
    EntityField,
    EntityType,
    EnumType,
    FloatType,
    IntegerType,
    LiteralType,
    OptionalType,
    RecursionHolder,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    UUIDType,
)
from ._meta import Meta, MetaStateKey
from ._type_utils import get_type_hints  # type: ignore[attr-defined]
from ._utils import get_attributes_doc, to_camelcase
from .metadata import (
    Alias,
    Discriminator,
    FieldFormat,
    Format,
    KeepDefaultForOptional,
    KeepNone,
    Max,
    MaxLength,
    Min,
    MinLength,
    NoFormat,
    NoneAsDefaultForOptional,
    NoneFormat,
    OmitNone,
)


if sys.version_info >= (3, 10):  # pragma: no cover
    from types import UnionType as StdUnionType
else:  # pragma: no cover
    StdUnionType = None

FrozenInstanceErrors: tuple[type[Exception], ...] = (dataclasses.FrozenInstanceError,)

try:
    import attr
    from attr.exceptions import FrozenInstanceError

    FrozenInstanceErrors += (FrozenInstanceError,)
except ImportError:  # pragma: no cover
    attr = None  # type: ignore
    FrozenInstanceError = None  # type: ignore


_NoneType = type(None)

_T = TypeVar('_T')


def describe_type(
    t: Any,
    meta: Optional[Meta] = None,
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]] = None,
) -> BaseType:
    args: tuple[Any, ...] = ()
    metadata = _get_annotated_metadata(t)
    type_repr = repr(t)
    if get_origin(t) == Annotated:  # unwrap annotated
        t = t.__origin__
    original_t = t
    if get_origin(t) in {Required, NotRequired}:  # unwrap TypedDict special forms
        t = t.__args__[0]
    if hasattr(t, '__origin__'):
        args = t.__args__
        t = t.__origin__
    # StdUnionType has no __origin__
    elif StdUnionType and isinstance(t, StdUnionType):  # type: ignore[truthy-function]
        args = t.__args__
        t = Union
    elif hasattr(t, '__parameters__'):
        # Если передан generic-класс без type-параметров, значит по PEP-484 заменяем все параметры на Any
        args = (Any,) * len(t.__parameters__)

    if not meta:
        meta = Meta(globals=_get_globals(t), state={})

    t = _evaluate_forwardref(t, meta)

    filed_format = _find_metadata(metadata, FieldFormat, NoFormat)
    none_format = _find_metadata(metadata, NoneFormat, KeepNone)
    none_as_default_for_optional = _find_metadata(metadata, NoneAsDefaultForOptional, KeepDefaultForOptional)
    custom_encoder = _find_metadata(metadata, CustomEncoder)
    annotation_wrapper = _wrap_annotated([filed_format, none_format, none_as_default_for_optional])

    meta_key = MetaStateKey(
        cls=original_t,
        field_format=filed_format,
        none_format=none_format,
        none_as_default_for_optional=none_as_default_for_optional,
    )

    if meta.has_in_state(meta_key):
        return RecursionHolder(
            name=_generate_name(original_t, filed_format, none_format, none_as_default_for_optional),
            state_key=meta_key,
            meta=meta,
            custom_encoder=None,
        )

    if custom_type_resolver and (custom_type := custom_type_resolver(t)):
        if custom_encoder is None:
            custom_encoder = CustomEncoder(
                serialize=custom_type.serialize,
                deserialize=custom_type.deserialize,
            )
        return CustomType(custom_encoder=custom_encoder, json_schema=custom_type.get_json_schema())

    if t is Any:
        return AnyType(custom_encoder=custom_encoder)

    if isinstance(t, type):
        simple_type_mapping: Mapping[type, type[BaseType]] = {
            bytes: BytesType,
            bool: BooleanType,
            date: DateType,
            time: TimeType,
            datetime: DateTimeType,
            UUID: UUIDType,
        }

        if simple := simple_type_mapping.get(t):
            return simple(custom_encoder=custom_encoder)

        number_type_mapping: Mapping[type, type[Union[IntegerType, FloatType]]] = {
            int: IntegerType,
            float: FloatType,
        }

        if number_type := number_type_mapping.get(t):
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return number_type(
                min=cast(Any, min_meta.value) if min_meta else None,
                max=cast(Any, max_meta.value) if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is Decimal:
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return DecimalType(
                min=min_meta.value if min_meta else None,
                max=max_meta.value if max_meta else None,
                custom_encoder=custom_encoder,
            )

        if t is str:
            min_length_meta = _find_metadata(metadata, MinLength)
            max_length_meta = _find_metadata(metadata, MaxLength)
            return StringType(
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
                custom_encoder=custom_encoder,
            )

        if t in {Sequence, list}:
            return ArrayType(
                item_type=(
                    describe_type(annotation_wrapper(args[0]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                custom_encoder=custom_encoder,
            )

        if t in {Mapping, dict}:
            return DictionaryType(
                key_type=(
                    describe_type(annotation_wrapper(args[0]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                value_type=(
                    describe_type(annotation_wrapper(args[1]), meta, custom_type_resolver)
                    if args
                    else AnyType(custom_encoder=None)
                ),
                omit_none=none_format.omit,
                custom_encoder=custom_encoder,
            )

        if t is tuple:
            if not args or Ellipsis in args:
                raise RuntimeError('Variable length tuples are not supported')
            return TupleType(
                item_types=[describe_type(annotation_wrapper(arg), meta, custom_type_resolver) for arg in args],
                custom_encoder=custom_encoder,
            )

        if issubclass(t, (Enum, IntEnum)):
            return EnumType(cls=t, items=list(t), custom_encoder=custom_encoder)

        if dataclasses.is_dataclass(t) or _is_attrs(t) or is_typeddict(t):
            meta.add_to_state(meta_key, None)
            entity_type = _describe_entity(
                t=t,
                original_t=original_t,
                cls_filed_format=filed_format,
                cls_none_format=none_format,
                custom_encoder=custom_encoder,
                cls_none_as_default_for_optional=none_as_default_for_optional,
                meta=meta,
                custom_type_resolver=custom_type_resolver,
            )
            meta.add_to_state(meta_key, entity_type)
            return entity_type

    if _is_literal_type(t):
        if args and all(isinstance(arg, (str, int)) for arg in args):
            return LiteralType(args=list(args), custom_encoder=custom_encoder)
        raise RuntimeError('Supported only Literal[str | int, ...]')

    if t is Union:
        if _NoneType in args:
            new_args = tuple(arg for arg in args if arg is not _NoneType)
            new_t = Union[new_args] if len(new_args) > 1 else new_args[0]  # type: ignore[unused-ignore]
            return OptionalType(
                inner=describe_type(annotation_wrapper(new_t), meta, custom_type_resolver),
                custom_encoder=None,
            )

        discriminator = _find_metadata(metadata, Discriminator)
        if not discriminator:
            return UnionType(
                item_types=[describe_type(annotation_wrapper(arg), meta, custom_type_resolver) for arg in args],
                union_repr=type_repr.removeprefix('typing.'),
                custom_encoder=custom_encoder,
            )

        if not all(dataclasses.is_dataclass(arg) or _is_attrs(arg) for arg in args):
            raise RuntimeError(
                f'Unions supported only for dataclasses or attrs. Provided: {t}[{",".join(map(str, args))}]'
            )

        meta = dataclasses.replace(meta, discriminator_field=discriminator.name)
        return DiscriminatedUnionType(
            item_types={
                _get_discriminator_value(arg, discriminator.name): describe_type(
                    annotation_wrapper(arg), meta, custom_type_resolver
                )
                for arg in args
            },
            dump_discriminator=discriminator.name,
            load_discriminator=_apply_format(filed_format, discriminator.name),
            custom_encoder=custom_encoder,
        )

    if isinstance(t, TypeVar):
        raise RuntimeError(f'Unfilled TypeVar: {t}')

    raise RuntimeError(f'Unknown type {t!r}')


@dataclasses.dataclass
class _Field(Generic[_T]):
    name: str
    type: type[_T]
    default: Union[DefaultValue[_T], DefaultValue[None]] = NOT_SET
    default_factory: Union[DefaultValue[Callable[[], _T]], DefaultValue[None]] = NOT_SET


def _describe_entity(
    t: Any,
    original_t: Any,
    cls_filed_format: FieldFormat,
    cls_none_format: NoneFormat,
    cls_none_as_default_for_optional: NoneAsDefaultForOptional,
    custom_encoder: Optional[CustomEncoder[Any, Any]],
    meta: Meta,
    custom_type_resolver: Optional[Callable[[Any], Optional[CustomTypeMeta[Any, Any]]]],
) -> Union[EntityType, TypedDictType]:
    # PEP-484: Replace all unfilled type parameters with Any
    if not hasattr(original_t, '__origin__') and getattr(original_t, '__parameters__', None):
        original_t = original_t[(Any,) * len(t.__parameters__)]

    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(original_t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}

    discriminator_field = meta.discriminator_field
    meta = dataclasses.replace(meta, globals=_get_globals(t), discriminator_field=None)

    fields = []
    for field in _get_entity_fields(t):
        type_ = types.get(field.name, field.type)
        type_ = Annotated[type_, cls_filed_format, cls_none_format, cls_none_as_default_for_optional]

        metadata = _get_annotated_metadata(type_)
        field_type = describe_type(type_, meta, custom_type_resolver)
        alias = _find_metadata(metadata, Alias)
        none_as_default_for_optional = _find_metadata(metadata, NoneAsDefaultForOptional)

        is_discriminator_field = field.name == discriminator_field
        required = not (field.default != NOT_SET or field.default_factory != NOT_SET) or is_discriminator_field

        default = field.default
        if (
            isinstance(field_type, OptionalType)
            and required
            and none_as_default_for_optional
            and none_as_default_for_optional.use
        ):
            default = DefaultValue.some(None)
            required = False

        fields.append(
            EntityField(
                name=field.name,
                dict_key=alias.value if alias else _apply_format(cls_filed_format, field.name),
                doc=docs.get(field.name),
                field_type=field_type,
                default=default,
                default_factory=field.default_factory,
                is_discriminator_field=is_discriminator_field,
                required=required,
            )
        )

    if is_typeddict(t):
        return TypedDictType(
            name=_generate_name(original_t, cls_filed_format, cls_none_format, cls_none_as_default_for_optional),
            fields=fields,
            omit_none=cls_none_format is OmitNone,
            doc=t.__doc__,
            custom_encoder=custom_encoder,
        )

    return EntityType(
        cls=t,
        name=_generate_name(original_t, cls_filed_format, cls_none_format, cls_none_as_default_for_optional),
        fields=fields,
        omit_none=cls_none_format is OmitNone,
        is_frozen=_is_frozen_dataclass(t, fields[0]) if fields else False,
        doc=_get_dataclass_doc(t),
        custom_encoder=custom_encoder,
    )


def _get_entity_fields(t: Any) -> Sequence[_Field[Any]]:
    if dataclasses.is_dataclass(t):
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(DefaultValue.some(f.default) if f.default is not dataclasses.MISSING else NOT_SET),
                default_factory=(
                    DefaultValue.some(f.default_factory) if f.default_factory is not dataclasses.MISSING else NOT_SET
                ),
            )
            for f in dataclasses.fields(t)
        ]
    if is_typeddict(t):
        return [
            _Field(
                name=field_name,
                type=field_type,
                default=NOT_SET if _is_required_in_typeddict(t, field_name) else DefaultValue.some(None),
                default_factory=NOT_SET,
            )
            for field_name, field_type in t.__annotations__.items()
        ]
    if _is_attrs(t):
        assert attr
        return [
            _Field(
                name=f.name,
                type=f.type,
                default=(
                    DefaultValue.some(f.default)
                    if (
                        f.default is not attr.NOTHING and not isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    )
                    else NOT_SET
                ),
                default_factory=(
                    DefaultValue.some(f.default.factory)  # pyright: ignore
                    if isinstance(f.default, attr.Factory)  # type: ignore[arg-type]
                    else NOT_SET
                ),
            )
            for f in attr.fields(t)
        ]

    raise RuntimeError(f"Unsupported type '{t}'")


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: _T) -> _T: ...


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: None = None) -> Optional[_T]: ...


def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: Optional[_T] = None) -> Optional[_T]:
    return next((ann for ann in annotations if isinstance(ann, type_)), default)


def _wrap_annotated(annotations: Iterable[Any]) -> Callable[[_T], _T]:
    def inner(type_: _T) -> _T:
        for ann in annotations:
            type_ = Annotated[type_, ann]  # type: ignore
        return type_

    return inner


def _get_annotated_metadata(t: Any) -> tuple[Any, ...]:
    if get_origin(t) == Annotated:
        return getattr(t, '__metadata__', ())
    return ()


def _apply_format(f: Optional[FieldFormat], value: str) -> str:
    if not f or f.format is Format.no_format:
        return value
    if f.format is Format.camel_case:
        return to_camelcase(value)
    assert_never(f.format)


def _generate_name(
    cls: Any,
    field_format: FieldFormat,
    none_format: NoneFormat,
    cls_none_as_default_for_optional: NoneAsDefaultForOptional,
) -> str:
    name = repr(cls).removeprefix("<class '").removesuffix("'>")
    nones = 'omit_nones' if none_format.omit else 'keep_nones'
    force_none = ',force_none' if cls_none_as_default_for_optional.use else ''
    return f'{name}[{field_format.format.value},{nones}{force_none}]'


def _get_globals(t: Any) -> dict[str, Any]:
    if t.__module__ in sys.modules:
        return sys.modules[t.__module__].__dict__.copy()
    return {}


def _evaluate_forwardref(t: type[_T], meta: Meta) -> type[_T]:
    if not isinstance(t, ForwardRef):
        return t

    return t._evaluate(meta.globals, {}, set())


def _get_discriminator_value(t: Any, name: str) -> str:
    fields = attr.fields(t) if attr and _is_attrs(t) else dataclasses.fields(t)
    for field in fields:
        if field.name == name:
            if _is_str_literal(field.type):
                args = get_args(field.type)
                if len(args) != 1:
                    raise RuntimeError(
                        f'Type {t} has invalid discriminator field "{name}". '
                        f'Discriminator supports only Literal[<str>] with one argument.'
                    )

                return cast(str, args[0])

            raise RuntimeError(
                f'Type {t} has invalid discriminator field "{name}" with type "{field.type!r}". '
                f'Discriminator supports only Literal[<str>].'
            )
    raise RuntimeError(f'Type {t} does not have discriminator field "{name}"')


def _is_str_literal(t: Any) -> bool:
    if _is_literal_type(t):
        args = get_args(t)
        if args and all(isinstance(arg, str) for arg in args):
            return True
    return False


def _is_literal_type(t: Any) -> bool:
    return t is Literal or get_origin(t) is Literal


def _is_attrs(t: Any) -> bool:
    return attr is not None and attr.has(t)


def _is_required_in_typeddict(t: Any, key: str) -> bool:
    if is_typeddict(t):
        if t.__total__:
            return key not in t.__optional_keys__
        return key in t.__required_keys__
    raise RuntimeError(f'Expected TypedDict, got "{t!r}"')


def _get_dataclass_doc(cls: Any) -> Optional[str]:
    """Dataclass has automatically generated docstring, which is not very useful."""
    doc: str = cls.__doc__
    if doc is None or doc.startswith(f'{cls.__name__}('):
        return None
    return doc


def _is_frozen_dataclass(cls: Any, field: EntityField) -> bool:
    try:
        obj = object.__new__(cls)
        setattr(obj, field.name, None)
    except FrozenInstanceErrors:
        return True
    else:
        return False



import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Generic, Literal, Optional, TypeVar, Union
from unittest import mock
from unittest.mock import ANY
from uuid import UUID

import attr
import pytest
from serpyco_rs._describe import (
    NOT_SET,
    AnyType,
    ArrayType,
    BooleanType,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue,
    DictionaryType,
    EntityField,
    EntityType,
    EnumType,
    FloatType,
    IntegerType,
    LiteralType,
    OptionalType,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    DiscriminatedUnionType,
    UUIDType,
    describe_type,
)
from serpyco_rs.metadata import Alias, CamelCase, Discriminator, Max, MaxLength, Min, MinLength, NoFormat
from typing_extensions import NotRequired, Required, TypedDict


T = TypeVar('T')
U = TypeVar('U')


def test_describe__dataclass__supported():
    class SomeEnum(Enum):
        a = 'a'

    @dataclass
    class SomeOtherEntity:
        a: int

    @dataclass
    class SomeEntity:
        """Doc"""

        a: int
        b: float
        c: Decimal
        d: bool
        e: str
        f: UUID
        g: time
        h: date
        i: datetime
        j: SomeEnum
        k: SomeOtherEntity
        l: list[int]
        m: Sequence[int]
        n: dict[str, int]
        o: Mapping[str, int]
        p: Any

    assert describe_type(SomeEntity) == EntityType(
        cls=SomeEntity,
        name=ANY,
        fields=[
            EntityField(
                name='a',
                field_type=IntegerType(min=None, max=None, custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='a',
            ),
            EntityField(
                name='b',
                field_type=FloatType(min=None, max=None, custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='b',
            ),
            EntityField(
                name='c',
                field_type=DecimalType(min=None, max=None, custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='c',
            ),
            EntityField(
                name='d',
                field_type=BooleanType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='d',
            ),
            EntityField(
                name='e',
                field_type=StringType(min_length=None, max_length=None, custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='e',
            ),
            EntityField(
                name='f',
                field_type=UUIDType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='f',
            ),
            EntityField(
                name='g',
                field_type=TimeType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='g',
            ),
            EntityField(
                name='h',
                field_type=DateType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='h',
            ),
            EntityField(
                name='i',
                field_type=DateTimeType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='i',
            ),
            EntityField(
                name='j',
                field_type=EnumType(cls=SomeEnum, items=[i for i in SomeEnum], custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='j',
            ),
            EntityField(
                name='k',
                field_type=EntityType(
                    cls=SomeOtherEntity,
                    name=ANY,
                    fields=[
                        EntityField(
                            name='a',
                            field_type=IntegerType(min=None, max=None, custom_encoder=None),
                            doc=None,
                            default=NOT_SET,
                            default_factory=NOT_SET,
                            dict_key='a',
                        )
                    ],
                    doc=None,
                    custom_encoder=None,
                ),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='k',
            ),
            EntityField(
                name='l',
                field_type=ArrayType(
                    item_type=IntegerType(min=None, max=None, custom_encoder=None),
                    custom_encoder=None,
                ),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='l',
            ),
            EntityField(
                name='m',
                field_type=ArrayType(
                    item_type=IntegerType(min=None, max=None, custom_encoder=None),
                    custom_encoder=None,
                ),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='m',
            ),
            EntityField(
                name='n',
                field_type=DictionaryType(
                    key_type=StringType(min_length=None, max_length=None, custom_encoder=None),
                    value_type=IntegerType(min=None, max=None, custom_encoder=None),
                    custom_encoder=None,
                ),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='n',
            ),
            EntityField(
                name='o',
                field_type=DictionaryType(
                    key_type=StringType(min_length=None, max_length=None, custom_encoder=None),
                    value_type=IntegerType(min=None, max=None, custom_encoder=None),
                    custom_encoder=None,
                ),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='o',
            ),
            EntityField(
                name='p',
                field_type=AnyType(custom_encoder=None),
                doc=None,
                default=NOT_SET,
                default_factory=NOT_SET,
                dict_key='p',
            ),
        ],
        doc='Doc',
        custom_encoder=None,
    )


def test_describe_dataclass__dict_type__works_without_type_parameters():
    @dataclass
    class SomeEntity:
        x: dict

    assert describe_type(SomeEntity).fields[0].field_type == DictionaryType(
        key_type=AnyType(custom_encoder=None),
        value_type=AnyType(custom_encoder=None),
        custom_encoder=None,
    )


def test_describe_dataclass__list_type__works_without_type_parameters():
    @dataclass
    class SomeEntity:
        x: list

    assert describe_type(SomeEntity).fields[0].field_type == ArrayType(
        item_type=AnyType(custom_encoder=None), custom_encoder=None
    )


def test_describe_dataclass__field_has_docstring__doc_filled():
    @dataclass
    class SomeEntity:
        x: int
        """Foo"""

    assert describe_type(SomeEntity).fields[0].doc == 'Foo'


def test_describe_dataclass__has_default__default_filled():
    @dataclass
    class SomeEntity:
        x: int = 1

    assert describe_type(SomeEntity).fields[0].default == DefaultValue.some(1)


def test_describe_dataclass__has_default_factory__default_factory_filled():
    def factory():
        return 1

    @dataclass
    class SomeEntity:
        x: int = field(default_factory=factory)

    assert describe_type(SomeEntity).fields[0].default == DefaultValue.none()
    assert describe_type(SomeEntity).fields[0].default_factory == DefaultValue.some(factory)


def test_describe_dataclass__generic_but_without_type_vars__filled_by_any():
    @dataclass
    class SomeEntity(Generic[T]):
        x: list[T]

    result: EntityType = describe_type(SomeEntity)
    assert result.fields[0].field_type == ArrayType(item_type=AnyType(custom_encoder=None), custom_encoder=None)


def test_describe_dataclass__generic_with_type_params__expected_right_type():
    @dataclass
    class SomeOtherEntity(Generic[T]):
        x: T

    @dataclass
    class SomeEntity(Generic[T]):
        x: list[T]
        y: SomeOtherEntity[T]

    result: EntityType = describe_type(SomeEntity[int])
    assert result.fields[0].field_type == ArrayType(item_type=IntegerType(custom_encoder=None), custom_encoder=None)
    assert result.fields[1].field_type == EntityType(
        cls=SomeOtherEntity,
        name=ANY,
        fields=[EntityField(name='x', field_type=IntegerType(custom_encoder=None), dict_key='x')],
        custom_encoder=None,
    )


def test_describe_dataclass__used_unknown_type_var__fail():
    @dataclass
    class SomeEntity(Generic[T]):
        x: list[T]
        y: U

    with pytest.raises(RuntimeError) as exc_info:
        describe_type(SomeEntity[int])

    assert exc_info.match('Unfilled TypeVar: ~U')


def test_describe__dataclass_with_forward_ref_annotation__parsed():
    @dataclass
    class SomeEntity:
        x: 'int'

    assert describe_type(SomeEntity).fields[0].field_type == IntegerType(custom_encoder=None)


def test_describe__dataclass_with_invalid_forward_ref_annotation__parsed():
    @dataclass
    class SomeEntity:
        x: 'intt'

    with pytest.raises(NameError) as exc_info:
        describe_type(SomeEntity)

    assert exc_info.match(r"name 'intt' is not defined")


def test_describe__dataclass_and_annotated_with_min_max__parsed():
    @dataclass
    class SomeEntity:
        x: Annotated[int, Min(10), Max(20)]

    result = describe_type(SomeEntity)

    assert result == EntityType(
        cls=SomeEntity,
        name=ANY,
        fields=[
            EntityField(
                name='x',
                dict_key='x',
                field_type=IntegerType(min=10, max=20, custom_encoder=None),
            )
        ],
        custom_encoder=None,
    )


def test_describe__dataclass_and_annotated_with_min_max_length__parsed():
    @dataclass
    class SomeEntity:
        x: Annotated[str, MinLength(10), MaxLength(20)]

    result = describe_type(SomeEntity)

    assert result == EntityType(
        cls=SomeEntity,
        name=ANY,
        doc=ANY,
        fields=[
            EntityField(
                name='x',
                dict_key='x',
                field_type=StringType(min_length=10, max_length=20, custom_encoder=None),
            )
        ],
        custom_encoder=None,
    )


def test_describe__attrs_and_field_has_docstring__doc_filled():
    @attr.s(auto_attribs=True)
    class SomeEntity:
        x: int
        """Foo"""

    assert describe_type(SomeEntity).fields[0].doc == 'Foo'


def test_describe__attrs_and_has_default__default_filled():
    @attr.s(auto_attribs=True)
    class SomeEntity:
        x: int = 1

    assert describe_type(SomeEntity).fields[0].default == DefaultValue.some(1)


def test_describe__attrs_and_has_default_factory__default_factory_filled():
    def factory():
        return 1

    @attr.s(auto_attribs=True)
    class SomeEntity:
        x: int = attr.ib(default=attr.Factory(factory))
        y: int = attr.ib(factory=factory)

    assert describe_type(SomeEntity).fields[0].default == NOT_SET
    assert describe_type(SomeEntity).fields[0].default_factory == DefaultValue.some(factory)
    assert describe_type(SomeEntity).fields[1].default == NOT_SET
    assert describe_type(SomeEntity).fields[1].default_factory == DefaultValue.some(factory)


def test_describe__attrs_with_forward_ref_annotation__parsed():
    @attr.s(auto_attribs=True)
    class SomeEntity:
        x: 'int'

    assert describe_type(SomeEntity).fields[0].field_type == IntegerType(custom_encoder=None)


def test_describe__attrs_with_invalid_forward_ref_annotation__parsed():
    @attr.s(auto_attribs=True)
    class SomeEntity:
        x: 'intt'

    with pytest.raises(NameError) as exc_info:
        describe_type(SomeEntity)

    assert exc_info.match(r"name 'intt' is not defined")


def test_describe__attrs_and_annotated_with_min_max__parsed():
    @attr.define
    class SomeEntity:
        x: Annotated[int, Min(10), Max(20)]

    result = describe_type(SomeEntity)

    assert result == EntityType(
        cls=SomeEntity,
        fields=[
            EntityField(
                name='x',
                dict_key='x',
                field_type=IntegerType(min=10, max=20, custom_encoder=None),
            )
        ],
        name=ANY,
        custom_encoder=None,
    )


def test_describe__attrs_and_annotated_with_min_max_length__parsed():
    @dataclass
    class SomeEntity:
        x: Annotated[str, MinLength(10), MaxLength(20)]

    result = describe_type(SomeEntity)

    assert result == EntityType(
        cls=SomeEntity,
        name=ANY,
        custom_encoder=None,
        fields=[
            EntityField(
                name='x',
                dict_key='x',
                field_type=StringType(min_length=10, max_length=20, custom_encoder=None),
            )
        ],
    )


def test_describe__type_with_typevar__fail():
    with pytest.raises(RuntimeError) as exc_info:
        describe_type(list[T])

    assert exc_info.match('Unfilled TypeVar: ~T')


def test_describe__unknown_type__fail():
    with pytest.raises(RuntimeError) as exc_info:
        describe_type(set[int])

    assert exc_info.match("Unknown type <class 'set'>")


def test_describe__optional__wrapped():
    assert describe_type(Optional[int]) == OptionalType(inner=IntegerType(custom_encoder=None), custom_encoder=None)


def test_describe__unions():
    assert describe_type(Union[int, str]) == UnionType(
        item_types=[IntegerType(custom_encoder=None), StringType(custom_encoder=None)],
        union_repr='Union[int, str]',
        custom_encoder=None,
    )


@pytest.mark.skipif(sys.version_info < (3, 10), reason='New style unions available after 3.10')
def test_describe__new_style_union_type__wrapped():
    assert describe_type(int | None) == OptionalType(inner=IntegerType(custom_encoder=None), custom_encoder=None)


def test_describe__tuple__parsed():
    assert describe_type(tuple[int, str]) == TupleType(
        item_types=[IntegerType(custom_encoder=None), StringType(custom_encoder=None)],
        custom_encoder=None,
    )


@pytest.mark.parametrize('t', [tuple, tuple[int, ...]])
def test_describe__invalid_tuple__error(t):
    with pytest.raises(RuntimeError) as exc_info:
        describe_type(t)

    assert exc_info.match('Variable length tuples are not supported')


def test_describe__dataclass_field_format__parsed():
    @dataclass
    class InnerEntity:
        foo_field: str
        bar_field: int

    @dataclass
    class Entity:
        inner_entity: Annotated[list[InnerEntity], NoFormat]
        some_filed: str

    assert describe_type(Annotated[Entity, CamelCase]) == EntityType(
        cls=Entity,
        name=ANY,
        custom_encoder=None,
        fields=[
            EntityField(
                name='inner_entity',
                dict_key='innerEntity',
                field_type=ArrayType(
                    custom_encoder=None,
                    item_type=EntityType(
                        name=ANY,
                        cls=InnerEntity,
                        fields=[
                            EntityField(
                                name='foo_field',
                                dict_key='foo_field',
                                field_type=StringType(custom_encoder=None),
                            ),
                            EntityField(
                                name='bar_field',
                                dict_key='bar_field',
                                field_type=IntegerType(custom_encoder=None),
                            ),
                        ],
                        doc=mock.ANY,
                        custom_encoder=None,
                    ),
                ),
            ),
            EntityField(
                name='some_filed',
                dict_key='someFiled',
                field_type=StringType(custom_encoder=None),
            ),
        ],
        doc=mock.ANY,
    )


def test_describe__literal():
    assert describe_type(Literal['foo', 'bar']) == LiteralType(args=['foo', 'bar'], custom_encoder=None)


def test_describe__tagged_union():
    @dataclass
    class Foo:
        val: int
        filed_type: Literal['foo']

    @dataclass
    class Bar:
        val: str
        filed_type: Literal['bar']

    @dataclass
    class Base:
        field: Annotated[Union[Foo, Bar], Discriminator('filed_type'), CamelCase]

    assert describe_type(Base) == EntityType(
        cls=Base,
        name=mock.ANY,
        fields=[
            EntityField(
                name='field',
                dict_key='field',
                field_type=DiscriminatedUnionType(
                    item_types={
                        'foo': EntityType(
                            cls=Foo,
                            name=mock.ANY,
                            fields=[
                                EntityField(
                                    name='val',
                                    dict_key='val',
                                    field_type=IntegerType(min=None, max=None, custom_encoder=None),
                                ),
                                EntityField(
                                    name='filed_type',
                                    dict_key='filedType',
                                    field_type=LiteralType(args=['foo'], custom_encoder=None),
                                    is_discriminator_field=True,
                                ),
                            ],
                            doc=mock.ANY,
                            custom_encoder=None,
                        ),
                        'bar': EntityType(
                            cls=Bar,
                            name=mock.ANY,
                            fields=[
                                EntityField(
                                    name='val',
                                    dict_key='val',
                                    field_type=StringType(min_length=None, max_length=None, custom_encoder=None),
                                ),
                                EntityField(
                                    name='filed_type',
                                    dict_key='filedType',
                                    field_type=LiteralType(args=['bar'], custom_encoder=None),
                                    is_discriminator_field=True,
                                ),
                            ],
                            doc=mock.ANY,
                            custom_encoder=None,
                        ),
                    },
                    load_discriminator='filedType',
                    dump_discriminator='filed_type',
                    custom_encoder=None,
                ),
            )
        ],
        doc=mock.ANY,
        custom_encoder=None,
    )


def test_describe__typed_dict():
    class Entity(TypedDict, Generic[T]):
        foo_filed: int
        bar_field: Annotated[NotRequired[str], Alias('barField')]
        generic_field: T

    assert describe_type(Entity[bool]) == TypedDictType(
        name=mock.ANY,
        fields=[
            EntityField(
                name='foo_filed',
                dict_key='foo_filed',
                field_type=IntegerType(custom_encoder=None),
            ),
            EntityField(
                name='bar_field',
                dict_key='barField',
                required=False,
                field_type=StringType(custom_encoder=None),
                default=DefaultValue.some(None),
            ),
            EntityField(
                name='generic_field',
                dict_key='generic_field',
                field_type=BooleanType(custom_encoder=None),
            ),
        ],
        custom_encoder=None,
    )


def test_describe__typed_dict__total_false():
    class Entity(TypedDict, total=False):
        foo: int
        bar: Required[str]

    assert describe_type(Entity[bool]) == TypedDictType(
        name=mock.ANY,
        fields=[
            EntityField(
                name='foo',
                dict_key='foo',
                required=False,
                field_type=IntegerType(custom_encoder=None),
                default=DefaultValue.some(None),
            ),
            EntityField(
                name='bar',
                dict_key='bar',
                field_type=StringType(custom_encoder=None),
            ),
        ],
        custom_encoder=None,
    )


def test_describe__dataclass__partial_typevars():
    @dataclass
    class Foo(Generic[T, U]):
        arr: list[T]
        key: U

    @dataclass
    class Bar(Foo[str, U]):
        pass

    assert describe_type(Bar) == EntityType(
        cls=Bar,
        name=ANY,
        fields=[
            EntityField(
                name='arr',
                dict_key='arr',
                field_type=ArrayType(item_type=StringType(custom_encoder=None), custom_encoder=None),
            ),
            EntityField(name='key', dict_key='key', field_type=AnyType(custom_encoder=None)),
        ],
        custom_encoder=None,
    )
    assert describe_type(Bar[bool]) == EntityType(
        cls=Bar,
        name=ANY,
        fields=[
            EntityField(
                name='arr',
                dict_key='arr',
                field_type=ArrayType(item_type=StringType(custom_encoder=None), custom_encoder=None),
            ),
            EntityField(name='key', dict_key='key', field_type=BooleanType(custom_encoder=None)),
        ],
        custom_encoder=None,
    )


def test_describe__dataclass__multiple_generic_inheritance():
    @dataclass
    class Foo(Generic[T, U]):
        arr: list[T]
        key: U

    @dataclass
    class Bar(Generic[T]):
        beer: T

    @dataclass
    class Baz(Foo[T, U], Bar[T], Generic[T, U]):
        pass

    @dataclass
    class Baz2(Foo[T, U], Bar[T]):
        pass

    assert describe_type(Baz[int, bool]) == EntityType(
        cls=Baz,
        name=ANY,
        fields=[
            EntityField(name='beer', dict_key='beer', field_type=IntegerType(custom_encoder=None)),
            EntityField(
                name='arr',
                dict_key='arr',
                field_type=ArrayType(item_type=IntegerType(custom_encoder=None), custom_encoder=None),
            ),
            EntityField(name='key', dict_key='key', field_type=BooleanType(custom_encoder=None)),
        ],
        custom_encoder=None,
    )
    assert describe_type(Baz2[int, bool]) == EntityType(
        cls=Baz2,
        name=ANY,
        fields=[
            EntityField(name='beer', dict_key='beer', field_type=IntegerType(custom_encoder=None)),
            EntityField(
                name='arr',
                dict_key='arr',
                field_type=ArrayType(item_type=IntegerType(custom_encoder=None), custom_encoder=None),
            ),
            EntityField(name='key', dict_key='key', field_type=BooleanType(custom_encoder=None)),
        ],
        custom_encoder=None,
    )
